package hu.unideb.ai.jars;

import java.util.ArrayList;
import java.util.List;

public class A {
	
	public static void main(String[] args) {
		search();
	}
	
	public static void search() {
		// DATABASE
		class Node {
			State state;
			Node parent;
			Operator creator;
			boolean open;
			double totalCost;
			
			public Node(State state, Node parent, Operator creator) {
				this.state = state;
				this.parent = parent;
				this.creator = creator;
				
				this.open = true;
				this.totalCost = parent == null ? state.heur() : 
					parent.totalCost - parent.state.heur() + creator.cost(parent.state) + state.heur();
			}
			
		}
		List<Node> nodes = new ArrayList<>();
		
		// Controller
		nodes.add( new Node(State.start(), null, null));
		
		while (true) {
			
			Node selected = null;
			for ( Node node : nodes )
				if ( node.open && ( selected == null || selected.totalCost > node.totalCost ) )
					selected = node;
			
			if ( selected == null ) {
				System.out.println("The problem is not solvable.");
				break;
			}
			
			if ( selected.state.isGoal() ) {
				System.out.println("Optimal solution!");
				System.out.println(selected.state);
				
				ArrayList<Operator> solution = new ArrayList<>();
				for ( Node n = selected; n.creator != null; n = n.parent )
					solution.add(0, n.creator);
				
				State temp = State.start();
				double cost = 0;
				for ( Operator o : solution ) {
					System.out.print(temp);
					System.out.print(" ");
					System.out.print(o);
					System.out.print(" ");
					cost += o.cost(temp);
					temp = o.apply(temp);
					System.out.print(temp);
					System.out.print(" ");
					System.out.println(cost);
					
				}
					
				break;
			}
			// Expand
			for ( Operator o : Operator.operators )
				if ( o.isApplicable(selected.state) ) {
					State newState = o.apply(selected.state);
					
					Node similar = null;
					for ( Node node:  nodes )
						if ( node.state.equals(newState) ) {
							similar = node;
							break;
						}
					
					if ( similar != null ) {
						if ( similar.totalCost > selected.totalCost - selected.state.heur() + o.cost(selected.state) + newState.heur()  ) {
							if ( similar.open ) {
								// Open node problem, update:
								similar.parent = selected;
								similar.creator = o;
								similar.totalCost = selected.totalCost - selected.state.heur() + o.cost(selected.state) + newState.heur();
							} else {
								// Closed node problem, update:
								similar.parent = selected;
								similar.creator = o;
								similar.totalCost = selected.totalCost - selected.state.heur() + o.cost(selected.state) + newState.heur();
								similar.open = true;
							}
						}
					} else {
						nodes.add( new Node(newState, selected, o));
					}
				}
			selected.open = false;
 			
		}
		
		
	}

}
