package hu.unideb.ai.jars;

import java.util.ArrayList;
import java.util.List;

public class OptimalSearch {
	
	public static void main(String[] args) {
		search();
	}
	
	public static void search() {
		// DATABASE
		class Node {
			State state;
			Node parent;
			Operator creator;
			boolean open;
			double pathCost;
			
			public Node(State state, Node parent, Operator creator) {
				this.state = state;
				this.parent = parent;
				this.creator = creator;
				
				this.open = true;
				this.pathCost = parent == null ? 0 : parent.pathCost + creator.cost(parent.state);
			}
			
		}
		List<Node> nodes = new ArrayList<>();
		
		// Controller
		nodes.add( new Node(State.start(), null, null));
		
		while (true) {
			
			Node selected = null;
			for ( Node node : nodes )
				if ( node.open && ( selected == null || selected.pathCost > node.pathCost ) )
					selected = node;
			
			if ( selected == null ) {
				System.out.println("The problem is not solvable.");
				break;
			}
			
			if ( selected.state.isGoal() ) {
				System.out.println("Optimal solution!");
				System.out.println(selected.state);
				
				ArrayList<Operator> solution = new ArrayList<>();
				for ( Node n = selected; n.creator != null; n = n.parent )
					solution.add(0, n.creator);
				
				State temp = State.start();
				double cost = 0;
				for ( Operator o : solution ) {
					System.out.print(temp);
					System.out.print(" ");
					System.out.print(o);
					System.out.print(" ");
					cost += o.cost(temp);
					temp = o.apply(temp);
					System.out.print(temp);
					System.out.print(" ");
					System.out.println(cost);
					
				}
					
				break;
			}
			// Expand
			for ( Operator o : Operator.operators )
				if ( o.isApplicable(selected.state) ) {
					State newState = o.apply(selected.state);
					
					Node similar = null;
					for ( Node node:  nodes )
						if ( node.state.equals(newState) ) {
							similar = node;
							break;
						}
					
					if ( similar != null ) {
						if ( similar.pathCost > selected.pathCost + o.cost(selected.state) ) {
							// Open node problem, update:
							similar.parent = selected;
							similar.creator = o;
							similar.pathCost = selected.pathCost + o.cost(selected.state);
						}
					} else {
						nodes.add( new Node(newState, selected, o));
					}
				}
			selected.open = false;
 			
		}
		
		
	}

}
